﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using System;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// TODO An operator for analyzing the best solution of Traveling Salesman Problems given in path representation using city coordinates.
  /// </summary>
  [Item("BestParameterConfigurationAnalyzer", "")]
  [StorableClass]
  public sealed class BestParameterConfigurationAnalyzer : SingleSuccessorOperator, IAnalyzer {

    public ScopeTreeLookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ScopeTreeLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<IRun> BestSolutionParameter {
      get { return (LookupParameter<IRun>)Parameters["BestSolution"]; }
    }
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public LookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public LookupParameter<IRun> BestKnownSolutionParameter {
      get { return (LookupParameter<IRun>)Parameters["BestKnownSolution"]; }
    }
    public LookupParameter<RunCollection> PopulationParameter {
      get { return (LookupParameter<RunCollection>)Parameters["Population"]; }
    }
    public LookupParameter<DoubleArray> ProblemQualityMediansParameter {
      get { return (LookupParameter<DoubleArray>)Parameters["ProblemQualityMedians"]; }
    }
    public LookupParameter<ConstrainedItemList<ISingleObjectiveProblem>> ProblemsParameter {
      get { return (LookupParameter<ConstrainedItemList<ISingleObjectiveProblem>>)Parameters[MetaOptimizationProblem.ProblemsParameterName]; }
    }

    public BestParameterConfigurationAnalyzer()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new LookupParameter<IRun>("BestSolution", ""));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", ""));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", ""));
      Parameters.Add(new LookupParameter<IRun>("BestKnownSolution", ""));
      Parameters.Add(new LookupParameter<RunCollection>("Population", ""));
      Parameters.Add(new LookupParameter<DoubleArray>("ProblemQualityMedians", ""));
      Parameters.Add(new LookupParameter<ConstrainedItemList<ISingleObjectiveProblem>>(MetaOptimizationProblem.ProblemsParameterName));
    }

    [StorableConstructor]
    private BestParameterConfigurationAnalyzer(bool deserializing) : base(deserializing) { }
    private BestParameterConfigurationAnalyzer(BestParameterConfigurationAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestParameterConfigurationAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      ResultCollection results = ResultsParameter.ActualValue;
      DoubleValue bestKnownQuality = BestKnownQualityParameter.ActualValue;
      ItemArray<ParameterConfigurationTree> parameterTrees = ParameterConfigurationParameter.ActualValue;

      int idxBest = qualities.Select((x, index) => new { index, x.Value }).OrderBy(x => x.Value).First().index;

      ParameterConfigurationTree best = (ParameterConfigurationTree)parameterTrees[idxBest];
      IRun bestRun = new Run();
      best.CollectResultValues(bestRun.Results);
      best.CollectParameterValues(bestRun.Parameters);

      if (bestKnownQuality == null || qualities[idxBest].Value < bestKnownQuality.Value) { // todo: respect Maximization:true/false
        BestKnownQualityParameter.ActualValue = new DoubleValue(qualities[idxBest].Value);

        BestKnownSolutionParameter.ActualValue = bestRun;
      }

      if (BestSolutionParameter.ActualValue == null) {
        BestSolutionParameter.ActualValue = bestRun;
        results.Add(new Result("Best Parameter Settings", bestRun));
      } else {
        BestSolutionParameter.ActualValue = bestRun;
        results["Best Parameter Settings"].Value = bestRun;
      }

      // population
      int i = 0;
      RunCollection rc = new RunCollection();
      foreach (ParameterConfigurationTree pt in parameterTrees.OrderBy(x => x.AverageQualityNormalized)) { // todo: respect Maximization:true/false
        IRun run = new Run();
        run.Name = string.Format("Individuum ({0})", i);
        pt.CollectResultValues(run.Results);
        pt.CollectParameterValues(run.Parameters);
        rc.Add(run);
        i++;
      }
      if (PopulationParameter.ActualValue == null) {
        PopulationParameter.ActualValue = rc;
        results.Add(new Result("Population", rc));
      } else {
        PopulationParameter.ActualValue = rc;
        results["Population"].Value = rc;
      }

      return base.Apply();
    }
  }
}
