﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using System.Collections.Generic;
using System;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// TODO An operator for analyzing the best solution of Traveling Salesman Problems given in path representation using city coordinates.
  /// </summary>
  [Item("MedianQualityAnalyzer", "")]
  [StorableClass]
  public sealed class MedianQualityAnalyzer : SingleSuccessorOperator, IAnalyzer {
    
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public ScopeTreeLookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ScopeTreeLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<DoubleArray> ProblemQualityMediansParameter {
      get { return (LookupParameter<DoubleArray>)Parameters["ProblemQualityMedians"]; }
    }
    public LookupParameter<ConstrainedItemList<ISingleObjectiveProblem>> ProblemsParameter {
      get { return (LookupParameter<ConstrainedItemList<ISingleObjectiveProblem>>)Parameters[MetaOptimizationProblem.ProblemsParameterName]; }
    }

    public MedianQualityAnalyzer()
      : base() {
        Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", ""));
      Parameters.Add(new ScopeTreeLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new LookupParameter<DoubleArray>("ProblemQualityMedians", ""));
      Parameters.Add(new LookupParameter<ConstrainedItemList<ISingleObjectiveProblem>>(MetaOptimizationProblem.ProblemsParameterName, ""));
    }

    [StorableConstructor]
    private MedianQualityAnalyzer(bool deserializing) : base(deserializing) { }
    private MedianQualityAnalyzer(MedianQualityAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new MedianQualityAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      ResultCollection results = ResultsParameter.ActualValue;
      ItemArray<ParameterConfigurationTree> solutions = ParameterConfigurationParameter.ActualValue;
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;

      if (ProblemQualityMediansParameter.ActualValue == null) {
        DoubleArray medians = CalculateMedians(solutions);
        ProblemQualityMediansParameter.ActualValue = medians;
        NormalizePopulation(solutions, qualities, medians);
        results.Add(new Result("ProblemsAverageQualityMedians", medians));
      }

      return base.Apply();
    }

    private DoubleArray CalculateMedians(ItemArray<ParameterConfigurationTree> solutions) {
      DoubleArray medians = new DoubleArray(ProblemsParameter.ActualValue.Count);
      for (int pi = 0; pi < ProblemsParameter.ActualValue.Count; pi++) {
        medians[pi] = solutions.Select(x => x.AverageQualities[pi]).Median();
      }
      return medians;
    }

    private void NormalizePopulation(ItemArray<ParameterConfigurationTree> solutions, ItemArray<DoubleValue> qualities, DoubleArray medians) {
      for (int i = 0; i < solutions.Length; i++) {
        qualities[i].Value = ParameterConfigurationEvaluator.NormalizeQualities(solutions[i], medians.ToArray());
      }
    }
  }
}
